@@ -1,9 +1,7 @@
 # gemato: CLI routines
-# vim:fileencoding=utf-8
-# (c) 2017-2020 Michał Górny
+# (c) 2017-2023 Michał Górny
 # Licensed under the terms of 2-clause BSD license
 
-from __future__ import print_function
 
 import argparse
 import datetime
@@ -15,7 +13,9 @@ import subprocess
 import sys
 import timeit
 
-from gemato.exceptions import GematoException
+from pathlib import Path
+
+from gemato.exceptions import GematoException, ManifestMismatch
 from gemato.find_top_level import find_top_level_manifest
 from gemato.hash import hash_file, hash_path
 from gemato.manifest import (
@@ -159,6 +159,19 @@ class VerifyingOpenPGPMixin(BaseOpenPGPMixin):
                                               keyserver=args.keyserver)
                 logging.info('Keys refreshed.')
 
+    def print_signatures(self, sigs):
+        for i, sig in enumerate(sigs):
+            if len(sigs) > 1:
+                logging.info(f"-- signature {i}")
+            logging.info(f"- status: {sig.sig_status}")
+            logging.info(f"- valid: {sig.valid_sig}, "
+                         f"trusted: {sig.trusted_sig}")
+            if sig.valid_sig:
+                logging.info("- primary key: "
+                             f"{sig.primary_key_fingerprint}")
+                logging.info(f"- subkey: {sig.fingerprint}")
+                logging.info(f"- timestamp: {sig.timestamp} UTC")
+
 
 class BaseManifestLoaderMixin:
     """
@@ -176,6 +189,16 @@ class BaseManifestLoaderMixin:
             '-x', '--one-file-system', action='store_true',
             help='Do not cross filesystem boundaries (report an error '
                  'instead)')
+        secugroup = subp.add_mutually_exclusive_group()
+        secugroup.add_argument(
+            "--require-secure-hashes", action="store_true",
+            default=None,
+            help="Require using secure hashes (default if Manifest is signed)")
+        secugroup.add_argument(
+            "--no-require-secure-hashes", action="store_false",
+            dest="require_secure_hashes",
+            help="Do not require using secure hashes (default if Manifest "
+                 "is not signed)")
 
     def parse_args(self, args, argp):
         super().parse_args(args, argp)
@@ -187,6 +210,9 @@ class BaseManifestLoaderMixin:
             self.init_kwargs['max_jobs'] = args.jobs
         if args.one_file_system:
             self.init_kwargs['allow_xdev'] = False
+        if args.require_secure_hashes is not None:
+            self.init_kwargs["require_secure_hashes"] = (
+                args.require_secure_hashes)
 
 
 class VerifyCommand(BaseManifestLoaderMixin, VerifyingOpenPGPMixin,
@@ -236,6 +262,8 @@ class VerifyCommand(BaseManifestLoaderMixin, VerifyingOpenPGPMixin,
                 logging.error(f'Top-level Manifest not found in {p}')
                 return 1
 
+            logging.info(f"Using top-level Manifest: {tlm}")
+
             start = timeit.default_timer()
             m = ManifestRecursiveLoader(tlm, **self.init_kwargs)
             if self.require_signed_manifest and not m.openpgp_signed:
@@ -249,22 +277,23 @@ class VerifyCommand(BaseManifestLoaderMixin, VerifyingOpenPGPMixin,
 
             if m.openpgp_signed:
                 logging.info('Valid OpenPGP signature found:')
-                logging.info(
-                    f'- primary key: '
-                    f'{m.openpgp_signature.primary_key_fingerprint}')
-                logging.info(
-                    f'- subkey: '
-                    f'{m.openpgp_signature.fingerprint}')
-                logging.info(
-                    f'- timestamp: '
-                    f'{m.openpgp_signature.timestamp} UTC')
+                self.print_signatures(m.openpgp_signature)
 
             logging.info(f'Verifying {p}...')
 
-            relpath = os.path.relpath(p, os.path.dirname(tlm))
+            tlmdir = os.path.dirname(tlm)
+            relpath = os.path.relpath(p, tlmdir)
             if relpath == '.':
                 relpath = ''
-            ret &= m.assert_directory_verifies(relpath, **self.kwargs)
+            try:
+                ret &= m.assert_directory_verifies(relpath, **self.kwargs)
+            except ManifestMismatch as e:
+                apparent_path = os.path.join(tlmdir, e.path)
+                real_path = os.path.realpath(apparent_path)
+                if apparent_path != real_path:
+                    logging.warning(
+                        f"Path contains symlinks. Real file path: {real_path}")
+                raise
 
             stop = timeit.default_timer()
             logging.info(f'{p} verified in {stop - start:.2f} seconds')
@@ -509,8 +538,7 @@ class HashCommand(GematoCommand):
             sz = h.pop('__size__')
             e = ManifestFileEntry(
                 p, sz,
-                dict((mh, h[hh]) for mh, hh
-                     in zip(self.hashes, hashlib_hashes)))
+                {mh: h[hh] for mh, hh in zip(self.hashes, hashlib_hashes)})
             print(' '.join(e.to_list('DATA' if p != '-' else 'STDIN')))
 
 
@@ -522,6 +550,12 @@ class OpenPGPVerifyCommand(VerifyingOpenPGPMixin, GematoCommand):
     def add_options(self, subp):
         super().add_options(subp)
 
+        subp.add_argument(
+            "--no-require-all-good",
+            dest="require_all_good",
+            action="store_false",
+            default=True,
+            help="Require only one good signature on multi-signature files")
         subp.add_argument(
             'paths', nargs='*', default=['-'],
             help='Paths to hash (defaults to "-" (stdin) if not '
@@ -531,6 +565,7 @@ class OpenPGPVerifyCommand(VerifyingOpenPGPMixin, GematoCommand):
         super().parse_args(args, argp)
 
         self.paths = args.paths
+        self.require_all_good = args.require_all_good
 
     def __call__(self):
         super().__call__()
@@ -541,11 +576,12 @@ class OpenPGPVerifyCommand(VerifyingOpenPGPMixin, GematoCommand):
             if p == '-':
                 f = sys.stdin
             else:
-                f = open(p, 'r')
+                f = open(p)
 
             try:
                 try:
-                    sig = self.openpgp_env.verify_file(f)
+                    sigs = self.openpgp_env.verify_file(
+                        f, require_all_good=self.require_all_good)
                 except GematoException as e:
                     logging.error(
                         f'OpenPGP verification failed for {p}:\n{e}')
@@ -553,10 +589,7 @@ class OpenPGPVerifyCommand(VerifyingOpenPGPMixin, GematoCommand):
                 else:
                     logging.info(
                         f'Valid OpenPGP signature found in {p}:')
-                    logging.info(
-                        f'- primary key: {sig.primary_key_fingerprint}')
-                    logging.info(f'- subkey: {sig.fingerprint}')
-                    logging.info(f'- timestamp: {sig.timestamp} UTC')
+                    self.print_signatures(sigs)
             finally:
                 if p != '-':
                     f.close()
@@ -564,6 +597,56 @@ class OpenPGPVerifyCommand(VerifyingOpenPGPMixin, GematoCommand):
         return 0 if ret else 1
 
 
+class OpenPGPVerifyDetachedCommand(VerifyingOpenPGPMixin, GematoCommand):
+    name = "openpgp-verify-detached"
+    help = "Verify a file against detached OpenPGP signature "
+
+    def add_options(self, subp):
+        super().add_options(subp)
+
+        subp.add_argument(
+            "--no-require-all-good",
+            dest="require_all_good",
+            action="store_false",
+            default=True,
+            help="Require only one good signature on multi-signature files")
+        subp.add_argument(
+            "signature_file",
+            type=Path,
+            help="Path to the file containing the OpenPGP signature")
+        subp.add_argument(
+            "data_file",
+            type=Path,
+            help="Path to the file to verify")
+
+    def parse_args(self, args, argp):
+        super().parse_args(args, argp)
+
+        self.signature_file = args.signature_file
+        self.data_file = args.data_file
+        self.require_all_good = args.require_all_good
+
+    def __call__(self):
+        super().__call__()
+
+        try:
+            sigs = self.openpgp_env.verify_detached(
+                self.signature_file, self.data_file,
+                require_all_good=self.require_all_good)
+        except GematoException as e:
+            logging.error(
+                f"OpenPGP verification failed for {self.data_file} "
+                f"(sig in {self.signature_file}):\n{e}")
+            return 1
+        else:
+            logging.info(
+                f"File {self.data_file} verified succesfully against "
+                f"the signature in {self.signature_file}:")
+            self.print_signatures(sigs)
+
+        return 0
+
+
 class GnuPGWrapCommand(VerifyingOpenPGPMixin, GematoCommand):
     name = 'gpg-wrap'
     help = ('Run specified command with GNUPGHOME set to OpenPGP '
@@ -610,6 +693,7 @@ def main(argv):
                 CreateCommand,
                 HashCommand,
                 OpenPGPVerifyCommand,
+                OpenPGPVerifyDetachedCommand,
                 GnuPGWrapCommand,
                 ]
     for cmdclass in commands:
@@ -635,5 +719,25 @@ def main(argv):
 
 
 def setuptools_main():
-    logging.getLogger().setLevel(logging.INFO)
+    try:
+        from rich.logging import RichHandler
+    except ImportError:
+        format_args = {
+            "format": "[{levelname:>7}] {message}",
+            "style": "{",
+        }
+    else:
+        format_args = {
+            "format": "{message}",
+            "style": "{",
+            "handlers": [RichHandler(show_time=False, show_path=False)],
+        }
+
+    logging.basicConfig(
+        level=logging.INFO,
+        **format_args)
     sys.exit(main(sys.argv))
+
+
+if __name__ == '__main__':
+    setuptools_main()
