@@ -1,11 +1,12 @@
 # gemato: OpenPGP verification support
-# vim:fileencoding=utf-8
-# (c) 2017-2020 Michał Górny
+# (c) 2017-2023 Michał Górny
 # Licensed under the terms of 2-clause BSD license
 
 import base64
+import dataclasses
 import datetime
 import email.utils
+import enum
 import errno
 import hashlib
 import logging
@@ -14,9 +15,12 @@ import os.path
 import shutil
 import subprocess
 import tempfile
+import typing
 import urllib.parse
 import warnings
 
+from pathlib import Path
+
 from gemato.exceptions import (
     OpenPGPNoImplementation,
     OpenPGPVerificationFailure,
@@ -45,16 +49,45 @@ GNUPG = os.environ.get('GNUPG', 'gpg')
 GNUPGCONF = os.environ.get('GNUPGCONF', 'gpgconf')
 
 
-class OpenPGPSignatureData:
-    __slots__ = ['fingerprint', 'timestamp', 'expire_timestamp',
-                 'primary_key_fingerprint']
+class OpenPGPSignatureStatus(enum.Enum):
+    GOOD = enum.auto()
+    BAD = enum.auto()
+    EXPIRED = enum.auto()
+    NO_PUBLIC_KEY = enum.auto()
+    ERROR = enum.auto()
+    EXPIRED_KEY = enum.auto()
+    REVOKED_KEY = enum.auto()
 
-    def __init__(self, fingerprint, timestamp, expire_timestamp,
-                 primary_key_fingerprint):
-        self.fingerprint = fingerprint
-        self.timestamp = timestamp
-        self.expire_timestamp = expire_timestamp
-        self.primary_key_fingerprint = primary_key_fingerprint
+
+@dataclasses.dataclass(order=True)
+class OpenPGPSignatureData:
+    fingerprint: str = ""
+    timestamp: typing.Optional[datetime.datetime] = None
+    expire_timestamp: typing.Optional[datetime.datetime] = None
+    primary_key_fingerprint: str = ""
+    sig_status: typing.Optional[OpenPGPSignatureStatus] = None
+    valid_sig: bool = False
+    trusted_sig: bool = False
+
+
+class OpenPGPSignatureList(list[OpenPGPSignatureData]):
+    # backwards compatibility with OpenPGPSignatureData
+
+    @property
+    def fingerprint(self) -> str:
+        return self[0].fingerprint
+
+    @property
+    def timestamp(self) -> typing.Optional[datetime.datetime]:
+        return self[0].timestamp
+
+    @property
+    def expire_timestamp(self) -> typing.Optional[datetime.datetime]:
+        return self[0].expire_timestamp
+
+    @property
+    def primary_key_fingerprint(self) -> str:
+        return self[0].primary_key_fingerprint
 
 
 ZBASE32_TRANSLATE = bytes.maketrans(
@@ -134,57 +167,146 @@ class SystemGPGEnvironment:
         else:
             return datetime.datetime.utcfromtimestamp(int(ts))
 
-    def verify_file(self, f):
+    def _process_gpg_verify_output(self,
+                                   out: bytes,
+                                   err: bytes,
+                                   require_all_good: bool,
+                                   ) -> OpenPGPSignatureList:
+        """Process the output of gpg --verify and return a siglist"""
+
+        sig_list = OpenPGPSignatureList()
+        for line in out.splitlines():
+            if line.startswith(b'[GNUPG:] NEWSIG'):
+                sig_list.append(OpenPGPSignatureData())
+            elif line.startswith(b'[GNUPG:] GOODSIG'):
+                assert sig_list and sig_list[-1].sig_status is None
+                sig_list[-1].sig_status = OpenPGPSignatureStatus.GOOD
+            elif line.startswith(b"[GNUPG:] BADSIG"):
+                assert sig_list and sig_list[-1].sig_status is None
+                sig_list[-1].sig_status = OpenPGPSignatureStatus.BAD
+            elif line.startswith(b"[GNUPG:] EXPSIG"):
+                assert sig_list and sig_list[-1].sig_status is None
+                sig_list[-1].sig_status = OpenPGPSignatureStatus.EXPIRED
+            elif line.startswith(b"[GNUPG:] ERRSIG"):
+                assert sig_list and sig_list[-1].sig_status is None
+                spl = line.split(b" ")
+                assert len(spl) >= 8
+                if spl[7] == b"9":
+                    sig_list[-1].sig_status = (
+                        OpenPGPSignatureStatus.NO_PUBLIC_KEY)
+                else:
+                    sig_list[-1].sig_status = OpenPGPSignatureStatus.ERROR
+            elif line.startswith(b'[GNUPG:] EXPKEYSIG'):
+                assert sig_list and sig_list[-1].sig_status is None
+                sig_list[-1].sig_status = OpenPGPSignatureStatus.EXPIRED_KEY
+            elif line.startswith(b'[GNUPG:] REVKEYSIG'):
+                assert sig_list and sig_list[-1].sig_status is None
+                sig_list[-1].sig_status = OpenPGPSignatureStatus.REVOKED_KEY
+            elif line.startswith(b'[GNUPG:] VALIDSIG'):
+                assert sig_list and not sig_list[-1].valid_sig
+                spl = line.split(b' ')
+                assert len(spl) >= 12
+                sig_list[-1].valid_sig = True
+                sig_list[-1].fingerprint = spl[2].decode('utf8')
+                sig_list[-1].timestamp = (
+                    self._parse_gpg_ts(spl[4].decode('utf8')))
+                sig_list[-1].expiration_timestamp = (
+                    self._parse_gpg_ts(spl[5].decode('utf8')))
+                sig_list[-1].primary_key_fingerprint = spl[11].decode('utf8')
+            elif line.startswith(b'[GNUPG:] TRUST_'):
+                assert sig_list
+                spl = line.split(b' ', 2)
+                if spl[1] in (b'TRUST_MARGINAL',
+                              b'TRUST_FULL',
+                              b'TRUST_ULTIMATE'):
+                    sig_list[-1].trusted_sig = True
+
+        if not sig_list:
+            raise OpenPGPUnknownSigFailure(
+                err.decode('utf8', errors='backslashreplace'))
+
+        # bad signature causes failure even without require_all_good
+        for sig in sig_list:
+            if sig.sig_status == OpenPGPSignatureStatus.BAD:
+                raise OpenPGPVerificationFailure(
+                    err.decode("utf8", errors="backslashreplace"), sig)
+
+        if not require_all_good:
+            if any(x.sig_status == OpenPGPSignatureStatus.GOOD and
+                   x.valid_sig and x.trusted_sig for x in sig_list):
+                return sig_list
+
+        for sig in sig_list:
+            if sig.sig_status == OpenPGPSignatureStatus.GOOD:
+                pass
+            elif sig.sig_status in (OpenPGPSignatureStatus.BAD,
+                                    OpenPGPSignatureStatus.EXPIRED,
+                                    OpenPGPSignatureStatus.NO_PUBLIC_KEY,
+                                    OpenPGPSignatureStatus.ERROR):
+                raise OpenPGPVerificationFailure(
+                    err.decode("utf8", errors="backslashreplace"), sig)
+            elif sig.sig_status == OpenPGPSignatureStatus.EXPIRED_KEY:
+                raise OpenPGPExpiredKeyFailure(
+                    err.decode('utf8', errors='backslashreplace'), sig)
+            elif sig.sig_status == OpenPGPSignatureStatus.REVOKED_KEY:
+                raise OpenPGPRevokedKeyFailure(
+                    err.decode('utf8', errors='backslashreplace'), sig)
+            else:
+                raise OpenPGPUnknownSigFailure(
+                    err.decode('utf8', errors='backslashreplace'), sig)
+
+            if not sig.valid_sig:
+                raise OpenPGPUnknownSigFailure(
+                    err.decode('utf8', errors='backslashreplace'), sig)
+            if not sig.trusted_sig:
+                raise OpenPGPUntrustedSigFailure(
+                    err.decode('utf8', errors='backslashreplace'), sig)
+
+        return sig_list
+
+    def verify_file(self,
+                    f: typing.IO[str],
+                    require_all_good: bool = True,
+                    ) -> OpenPGPSignatureList:
         """
         Perform an OpenPGP verification of Manifest data in open file @f.
         The file should be open in text mode and set at the beginning
         (or start of signed part). Raises an exception if the verification
         fails.
+
+        If require_all_good is True and the file contains multiple OpenPGP
+        signatures, all signatures have to be good and trusted in order
+        for the verification to succeed. Otherwise, a single good signature
+        is considered sufficient.
         """
 
         exitst, out, err = self._spawn_gpg(
             [GNUPG, '--batch', '--status-fd', '1', '--verify'],
-            f.read().encode('utf8'),
-            raise_on_error=OpenPGPVerificationFailure)
+            f.read().encode('utf8'))
+        return self._process_gpg_verify_output(out, err, require_all_good)
 
-        is_good = False
-        is_trusted = False
-        sig_data = None
+    def verify_detached(self,
+                        signature_file: Path,
+                        data_file: Path,
+                        require_all_good: bool = True,
+                        ) -> OpenPGPSignatureList:
+        """
+        Verify the file against a detached signature
 
-        # process the output of gpg to find the exact result
-        for line in out.splitlines():
-            if line.startswith(b'[GNUPG:] GOODSIG'):
-                is_good = True
-            elif line.startswith(b'[GNUPG:] EXPKEYSIG'):
-                raise OpenPGPExpiredKeyFailure(
-                    err.decode('utf8', errors='backslashreplace'))
-            elif line.startswith(b'[GNUPG:] REVKEYSIG'):
-                raise OpenPGPRevokedKeyFailure(
-                    err.decode('utf8', errors='backslashreplace'))
-            elif line.startswith(b'[GNUPG:] VALIDSIG'):
-                spl = line.split(b' ')
-                assert len(spl) >= 12
-                fp = spl[2].decode('utf8')
-                ts = self._parse_gpg_ts(spl[4].decode('utf8'))
-                expts = self._parse_gpg_ts(spl[5].decode('utf8'))
-                pkfp = spl[11].decode('utf8')
+        Verify the data from data_file against the detached signature
+        from signature_file. Both files are specified by Path.
+        Raise an exception if the verification fails.
 
-                sig_data = OpenPGPSignatureData(fp, ts, expts, pkfp)
-            elif line.startswith(b'[GNUPG:] TRUST_'):
-                spl = line.split(b' ', 2)
-                if spl[1] in (b'TRUST_MARGINAL',
-                              b'TRUST_FULL',
-                              b'TRUST_ULTIMATE'):
-                    is_trusted = True
+        If require_all_good is True and the file contains multiple OpenPGP
+        signatures, all signatures have to be good and trusted in order
+        for the verification to succeed. Otherwise, a single good signature
+        is considered sufficient.
+        """
 
-        # require both GOODSIG and VALIDSIG
-        if not is_good or sig_data is None:
-            raise OpenPGPUnknownSigFailure(
-                err.decode('utf8', errors='backslashreplace'))
-        if not is_trusted:
-            raise OpenPGPUntrustedSigFailure(
-                err.decode('utf8', errors='backslashreplace'))
-        return sig_data
+        _, out, err = self._spawn_gpg(
+            [GNUPG, "--batch", "--status-fd", "1", "--verify",
+             str(signature_file), str(data_file)])
+        return self._process_gpg_verify_output(out, err, require_all_good)
 
     def clear_sign_file(self, f, outf, keyid=None):
         """
@@ -229,6 +351,19 @@ class SystemGPGEnvironment:
         return (p.wait(), out, err)
 
 
+def _rmtree_error_handler(func, path, exc_info):
+    # ignore ENOENT -- it probably means a race condition between
+    # us and gpg-agent cleaning up after itself
+    # also non-empty directory due to races, and EBUSY for NFS:
+    # https://bugs.gentoo.org/684172
+    if (not isinstance(exc_info[1], OSError)
+            or exc_info[1].errno not in (errno.ENOENT,
+                                         errno.ENOTEMPTY,
+                                         errno.EEXIST,
+                                         errno.EBUSY)):
+        raise exc_info[1]
+
+
 class IsolatedGPGEnvironment(SystemGPGEnvironment):
     """
     An isolated environment for OpenPGP routines. Used to get reliable
@@ -279,19 +414,6 @@ debug-level guru
     def clone(self):
         return IsolatedGPGEnvironment(debug=self.debug, proxy=self.proxy)
 
-    @staticmethod
-    def _rmtree_error_handler(func, path, exc_info):
-        # ignore ENOENT -- it probably means a race condition between
-        # us and gpg-agent cleaning up after itself
-        # also non-empty directory due to races, and EBUSY for NFS:
-        # https://bugs.gentoo.org/684172
-        if (not isinstance(exc_info[1], OSError)
-                or exc_info[1].errno not in (errno.ENOENT,
-                                             errno.ENOTEMPTY,
-                                             errno.EEXIST,
-                                             errno.EBUSY)):
-            raise exc_info[1]
-
     def close(self):
         if self._home is not None:
             ret, sout, serr = self._spawn_gpg(
@@ -304,7 +426,7 @@ debug-level guru
                 # we need to loop due to ENOTEMPTY potential
                 while os.path.isdir(self._home):
                     shutil.rmtree(self._home,
-                                  onerror=self._rmtree_error_handler)
+                                  onerror=_rmtree_error_handler)
             else:
                 logging.debug(f'GNUPGHOME left for debug purposes: '
                               f'{self._home}')
@@ -416,10 +538,14 @@ debug-level guru
             }
         for a in addrs:
             url = get_wkd_url(a)
-            resp = requests.get(url, proxies=proxies)
-            if resp.status_code != 200:
-                logging.debug(f'refresh_keys_wkd(): failing due to failed'
-                              f'request for {url}: {resp}')
+            try:
+                resp = requests.get(url, proxies=proxies)
+                resp.raise_for_status()
+            except (requests.exceptions.ConnectionError,
+                    requests.exceptions.HTTPError,
+                    ) as e:
+                logging.debug(f'refresh_keys_wkd(): failing due to failed '
+                              f'request for {url}: {e}')
                 return False
             data += resp.content
 
