@@ -1,10 +1,12 @@
 # gemato: OpenPGP signature support tests
-# vim:fileencoding=utf-8
-# (c) 2017-2020 Michał Górny
+# (c) 2017-2023 Michał Górny
 # Licensed under the terms of 2-clause BSD license
 
+import base64
+import contextlib
 import datetime
 import io
+import logging
 import os
 import shlex
 import signal
@@ -25,6 +27,7 @@ from gemato.exceptions import (
     OpenPGPKeyRefreshError,
     OpenPGPRuntimeError,
     OpenPGPUntrustedSigFailure,
+    ManifestInsecureHashes,
     )
 from gemato.manifest import ManifestFile
 from gemato.openpgp import (
@@ -32,29 +35,26 @@ from gemato.openpgp import (
     IsolatedGPGEnvironment,
     PGPyEnvironment,
     get_wkd_url,
+    OpenPGPSignatureList,
+    OpenPGPSignatureData,
+    OpenPGPSignatureStatus,
     )
 from gemato.recursiveloader import ManifestRecursiveLoader
 
 from tests.keydata import (
-    PUBLIC_KEY, SECRET_KEY, PUBLIC_SUBKEY, UID,
-    UID_NOEMAIL, PUBLIC_KEY_NOEMAIL_SIG,
-    UID_NONUTF, PUBLIC_KEY_NONUTF_SIG,
-    PUBLIC_KEY_SIG, PUBLIC_SUBKEY_SIG, EXPIRED_KEY_SIG, REVOCATION_SIG,
-    OTHER_PUBLIC_KEY, OTHER_PUBLIC_KEY_UID, OTHER_PUBLIC_KEY_SIG,
-    UNEXPIRE_SIG,
+    PRIVATE_KEY_ID, KEY_FINGERPRINT, OTHER_KEY_FINGERPRINT, VALID_PUBLIC_KEY,
+    VALID_KEY_NOEMAIL, VALID_KEY_NONUTF, COMBINED_PUBLIC_KEYS,
+    VALID_KEY_SUBKEY, PRIVATE_KEY, EXPIRED_PUBLIC_KEY, REVOKED_PUBLIC_KEY,
+    OTHER_VALID_PUBLIC_KEY, UNSIGNED_PUBLIC_KEY, FORGED_PUBLIC_KEY,
+    UNSIGNED_SUBKEY, FORGED_SUBKEY, SIG_TIMESTAMP, SUBKEY_FINGERPRINT,
+    SUBKEY_SIG_TIMESTAMP, UNEXPIRE_PUBLIC_KEY, OLD_UNEXPIRE_PUBLIC_KEY,
+    FORGED_UNEXPIRE_KEY, TWO_SIGNATURE_PUBLIC_KEYS, SECOND_KEY_FINGERPRINT,
+    SECOND_VALID_PUBLIC_KEY, TWO_KEYS_ONE_EXPIRED,
     )
+from tests.test_recursiveloader import INSECURE_HASH_TESTS
 from tests.testutil import HKPServer
 
 
-VALID_PUBLIC_KEY = PUBLIC_KEY + UID + PUBLIC_KEY_SIG
-EXPIRED_PUBLIC_KEY = PUBLIC_KEY + UID + EXPIRED_KEY_SIG
-REVOKED_PUBLIC_KEY = PUBLIC_KEY + REVOCATION_SIG + UID + PUBLIC_KEY_SIG
-OLD_UNEXPIRE_PUBLIC_KEY = PUBLIC_KEY + UID + PUBLIC_KEY_SIG
-UNEXPIRE_PUBLIC_KEY = PUBLIC_KEY + UID + UNEXPIRE_SIG
-
-PRIVATE_KEY = SECRET_KEY + UID + PUBLIC_KEY_SIG
-PRIVATE_KEY_ID = b'0x136880E72A7B1384'
-
 MALFORMED_PUBLIC_KEY = b'''
 -----BEGIN PGP PUBLIC KEY BLOCK-----
 
@@ -75,10 +75,7 @@ jCvJNJ7pU8YnJSRTQDH0PZEupAdzDU/AhGSrBz5+Jr7N0pQIxq4duE/Q
 -----END PGP PUBLIC KEY BLOCK-----
 '''
 
-SIGNED_MANIFEST = u'''
------BEGIN PGP SIGNED MESSAGE-----
-Hash: SHA256
-
+COMMON_MANIFEST_TEXT = """\
 TIMESTAMP 2017-10-22T18:06:41Z
 MANIFEST eclass/Manifest 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
  SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
@@ -88,7 +85,14 @@ DATA myebuild-0.ebuild 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
 MISC metadata.xml 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
  SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
 DIST mydistfile.tar.gz 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
- SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
+ SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709\
+"""
+
+SIGNED_MANIFEST = f"""
+-----BEGIN PGP SIGNED MESSAGE-----
+Hash: SHA256
+
+{COMMON_MANIFEST_TEXT}
 -----BEGIN PGP SIGNATURE-----
 
 iQEzBAEBCAAdFiEEgeEsFr2NzWC+GAhFE2iA5yp7E4QFAloCx+YACgkQE2iA5yp7
@@ -100,9 +104,9 @@ ZIfqeaQ4NoefmxQunyEjT+8X2DMaEeHQni7dwjQc+FiN4ReV9aWbLo2O2cArqEHR
 mkkhTd2Auao4D2K74BePBuiZ9+eDQA==
 =khff
 -----END PGP SIGNATURE-----
-'''
+"""
 
-DASH_ESCAPED_SIGNED_MANIFEST = u'''
+DASH_ESCAPED_SIGNED_MANIFEST = '''
 -----BEGIN PGP SIGNED MESSAGE-----
 Hash: SHA256
 
@@ -129,7 +133,7 @@ mkkhTd2Auao4D2K74BePBuiZ9+eDQA==
 -----END PGP SIGNATURE-----
 '''
 
-MODIFIED_SIGNED_MANIFEST = u'''
+MODIFIED_SIGNED_MANIFEST = '''
 -----BEGIN PGP SIGNED MESSAGE-----
 Hash: SHA256
 
@@ -155,20 +159,11 @@ mkkhTd2Auao4D2K74BePBuiZ9+eDQA==
 -----END PGP SIGNATURE-----
 '''
 
-EXPIRED_SIGNED_MANIFEST = u'''
+EXPIRED_SIGNED_MANIFEST = f"""
 -----BEGIN PGP SIGNED MESSAGE-----
 Hash: SHA256
 
-TIMESTAMP 2017-10-22T18:06:41Z
-MANIFEST eclass/Manifest 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
- SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
-IGNORE local
-DATA myebuild-0.ebuild 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
- SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
-MISC metadata.xml 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
- SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
-DIST mydistfile.tar.gz 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
- SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
+{COMMON_MANIFEST_TEXT}
 -----BEGIN PGP SIGNATURE-----
 
 iQE5BAEBCAAjFiEEgeEsFr2NzWC+GAhFE2iA5yp7E4QFAlnxCXcFgwABUYAACgkQ
@@ -180,37 +175,13 @@ HSxRUCkTK1jBuP/v/rjdqUJw3LXAbjxFl9SyUX4AgCgHqgso3IZwjAprQRKNSObO
 t5pTRGhLWgdLUrs7vRB7wf7F8h4sci/YBKJRFA==
 =VGMV
 -----END PGP SIGNATURE-----
-'''
+"""
 
-KEY_FINGERPRINT = '81E12C16BD8DCD60BE180845136880E72A7B1384'
-SIG_TIMESTAMP = datetime.datetime(2017, 11, 8, 9, 1, 26)
-
-OTHER_VALID_PUBLIC_KEY = (OTHER_PUBLIC_KEY + OTHER_PUBLIC_KEY_UID +
-                          OTHER_PUBLIC_KEY_SIG)
-OTHER_KEY_FINGERPRINT = '4B8349B90C56EE7F054D52871822F5424EB6DA81'
-
-VALID_KEY_NOEMAIL = PUBLIC_KEY + UID_NOEMAIL + PUBLIC_KEY_NOEMAIL_SIG
-VALID_KEY_NONUTF = PUBLIC_KEY + UID_NONUTF + PUBLIC_KEY_NONUTF_SIG
-
-VALID_KEY_SUBKEY = (PUBLIC_KEY + UID + PUBLIC_KEY_SIG + PUBLIC_SUBKEY +
-                    PUBLIC_SUBKEY_SIG)
-SUBKEY_FINGERPRINT = '7E9DDE3CBE47E437418DF74038B9D2F76CC833CC'
-SUBKEY_SIG_TIMESTAMP = datetime.datetime(2020, 8, 25, 12, 40, 12)
-
-SUBKEY_SIGNED_MANIFEST = u'''
+SUBKEY_SIGNED_MANIFEST = f"""
 -----BEGIN PGP SIGNED MESSAGE-----
 Hash: SHA256
 
-TIMESTAMP 2017-10-22T18:06:41Z
-MANIFEST eclass/Manifest 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
- SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
-IGNORE local
-DATA myebuild-0.ebuild 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
- SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
-MISC metadata.xml 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
- SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
-DIST mydistfile.tar.gz 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
- SHA1 da39a3ee5e6b4b0d3255bfef95601890afd80709
+{COMMON_MANIFEST_TEXT}
 -----BEGIN PGP SIGNATURE-----
 
 iLMEAQEIAB0WIQR+nd48vkfkN0GN90A4udL3bMgzzAUCX0UGrAAKCRA4udL3bMgz
@@ -219,24 +190,32 @@ g7VRUc//lz5sDUShdUUlTYjCPGLaYf2rBZHqd5POGJOsbzu1Tmtd8uhWFWnl8Kip
 n4XmpdPvu+UdAHpQIGzKoNOEDJpZ5CzPLhYa5KgZiJhpYsDXgg==
 =lpJi
 -----END PGP SIGNATURE-----
-'''
+"""
 
+TWO_SIGNATURES = """
+iQFHBAABCAAxFiEEgeEsFr2NzWC+GAhFE2iA5yp7E4QFAmPMHYQTHGdlbWF0b0Bl
+eGFtcGxlLmNvbQAKCRATaIDnKnsThCDWB/95B9njv423M94uRdpPqSNqTpAokNhy
+V0hjnhpiqnY85iFdL1Zc/rvhuxYbZezrig3dqctLseWYcx2mINBTLZqWHk5/NKEm
+rd8iCdXZU1B7yo/HCfzUYR4HX5wISCiRjKimFFgkWKOg7KYGOqqrwLjAjaYJKmL5
+L7R5joHpGbp87jix7c0ruSIMslQg5PbJ6/YAQWyOPTcZvqMFieJ8tqE/G2FabQcs
+YRHEGu1x8wNY40rFzWd90ICR/hPjXZlCdCN2qk7hs+Coasb29n6pXjmt5L8/ICcL
+zApRg8cetid6/SIzUSwiVqBt7i8noYWbgaazNt3HDlGq55v21dkOhmrXiIkEABYI
+ADEWIQR1jj6cjPscaH2bJCVTcI9ps0i0zAUCY8wd6BMcc2Vjb25kQGV4YW1wbGUu
+Y29tAAoJEFNwj2mzSLTMHKcA/0QbVl3PafYp45PFFo2e/knGKJKrm8D4bUH9wS5h
+dchVAP0RSzkUQPP7Zs+2uHQItkqbXJyrBBHOqjGzeh39sWVuAw==
+"""
 
-def break_sig(sig):
-    """Return signature packet mangled to mismatch the signed key"""
-    return sig[:-1] + b'\x55'
+TWO_SIGNATURE_MANIFEST = f"""
+-----BEGIN PGP SIGNED MESSAGE-----
+Hash: SHA256
 
+{COMMON_MANIFEST_TEXT}
+-----BEGIN PGP SIGNATURE-----
 
-FORGED_PUBLIC_KEY = PUBLIC_KEY + UID + break_sig(PUBLIC_KEY_SIG)
-FORGED_SUBKEY = (PUBLIC_KEY + UID + PUBLIC_KEY_SIG + PUBLIC_SUBKEY +
-                 break_sig(PUBLIC_SUBKEY_SIG))
-FORGED_UNEXPIRE_KEY = (PUBLIC_KEY + UID + EXPIRED_KEY_SIG +
-                       break_sig(UNEXPIRE_SIG))
-
-UNSIGNED_PUBLIC_KEY = PUBLIC_KEY + UID
-UNSIGNED_SUBKEY = PUBLIC_KEY + UID + PUBLIC_KEY_SIG + PUBLIC_SUBKEY
-
-COMBINED_PUBLIC_KEYS = OTHER_VALID_PUBLIC_KEY + VALID_PUBLIC_KEY
+{TWO_SIGNATURES}
+=wG4b
+-----END PGP SIGNATURE-----
+"""
 
 
 def strip_openpgp(text):
@@ -257,6 +236,25 @@ MANIFESTS_BAD_SIG = [
 ]
 
 
+# workaround pyflakes' limitations
+_ = COMBINED_PUBLIC_KEYS
+_ = EXPIRED_PUBLIC_KEY
+_ = FORGED_PUBLIC_KEY
+_ = FORGED_SUBKEY
+_ = FORGED_UNEXPIRE_KEY
+_ = OLD_UNEXPIRE_PUBLIC_KEY
+_ = OTHER_VALID_PUBLIC_KEY
+_ = SECOND_VALID_PUBLIC_KEY
+_ = TWO_KEYS_ONE_EXPIRED
+_ = TWO_SIGNATURE_PUBLIC_KEYS
+_ = UNEXPIRE_PUBLIC_KEY
+_ = UNSIGNED_PUBLIC_KEY
+_ = UNSIGNED_SUBKEY
+_ = VALID_KEY_NOEMAIL
+_ = VALID_KEY_NONUTF
+_ = VALID_KEY_SUBKEY
+
+
 @pytest.mark.parametrize('manifest_var',
                          MANIFESTS_GOOD_SIG + MANIFESTS_BAD_SIG)
 def test_noverify_goodish_manifest_load(manifest_var):
@@ -313,7 +311,7 @@ def test_noverify_recursive_manifest_loader(tmp_path, write_back):
 
     if write_back:
         m.save_manifest('Manifest')
-        with open(tmp_path / 'Manifest', 'r') as f:
+        with open(tmp_path / 'Manifest') as f:
             assert f.read() == strip_openpgp(MODIFIED_SIGNED_MANIFEST)
 
 
@@ -337,6 +335,7 @@ class MockedSystemGPGEnvironment(SystemGPGEnvironment):
     """System environment variant mocked to use isolated GNUPGHOME"""
     def __init__(self, *args, **kwargs):
         self._tmpdir = tempfile.TemporaryDirectory()
+        self._home = self._tmpdir.name
         os.environ['GNUPGHOME'] = self._tmpdir.name
         super().__init__(*args, **kwargs)
 
@@ -348,6 +347,9 @@ class MockedSystemGPGEnvironment(SystemGPGEnvironment):
 
     def close(self):
         if self._tmpdir is not None:
+            IsolatedGPGEnvironment.close(self)
+            # we need to recreate it to make cleanup() happy
+            os.mkdir(self._tmpdir.name)
             self._tmpdir.cleanup()
             self._tmpdir = None
             os.environ.pop('GNUPGHOME', None)
@@ -388,6 +390,13 @@ MANIFEST_VARIANTS = [
     ('SIGNED_MANIFEST', 'COMBINED_PUBLIC_KEYS', None),
     ('DASH_ESCAPED_SIGNED_MANIFEST', 'VALID_PUBLIC_KEY', None),
     ('SUBKEY_SIGNED_MANIFEST', 'VALID_KEY_SUBKEY', None),
+    # == Manifest with two signatures ==
+    ("TWO_SIGNATURE_MANIFEST", "TWO_SIGNATURE_PUBLIC_KEYS", None),
+    ("TWO_SIGNATURE_MANIFEST", "VALID_PUBLIC_KEY", OpenPGPVerificationFailure),
+    ("TWO_SIGNATURE_MANIFEST", "SECOND_VALID_PUBLIC_KEY",
+     OpenPGPVerificationFailure),
+    ("TWO_SIGNATURE_MANIFEST", "TWO_KEYS_ONE_EXPIRED",
+     OpenPGPExpiredKeyFailure),
     # == using private key ==
     ('SIGNED_MANIFEST', 'PRIVATE_KEY', None),
     # == bad manifests ==
@@ -415,14 +424,40 @@ MANIFEST_VARIANTS = [
 ]
 
 
-def assert_signature(sig, manifest_var):
+def assert_signature(sig: OpenPGPSignatureList,
+                     manifest_var: str,
+                     expect_both: bool = True,
+                     ) -> None:
     """Make assertions about the signature"""
-    if manifest_var == 'SUBKEY_SIGNED_MANIFEST':
+    if manifest_var == "TWO_SIGNATURE_MANIFEST":
+        no_key_sig = OpenPGPSignatureData(
+            sig_status=OpenPGPSignatureStatus.NO_PUBLIC_KEY)
+        assert sorted(sig) == sorted([
+            OpenPGPSignatureData(
+                fingerprint=KEY_FINGERPRINT,
+                timestamp=datetime.datetime(2023, 1, 21, 17, 14, 44),
+                primary_key_fingerprint=KEY_FINGERPRINT,
+                sig_status=OpenPGPSignatureStatus.GOOD,
+                trusted_sig=True,
+                valid_sig=True,
+                ),
+            OpenPGPSignatureData(
+                fingerprint=SECOND_KEY_FINGERPRINT,
+                timestamp=datetime.datetime(2023, 1, 21, 17, 16, 24),
+                primary_key_fingerprint=SECOND_KEY_FINGERPRINT,
+                sig_status=OpenPGPSignatureStatus.GOOD,
+                trusted_sig=True,
+                valid_sig=True,
+                ) if expect_both else no_key_sig,
+        ])
+    elif manifest_var == 'SUBKEY_SIGNED_MANIFEST':
+        assert len(sig) == 1
         assert sig.fingerprint == SUBKEY_FINGERPRINT
         assert sig.timestamp == SUBKEY_SIG_TIMESTAMP
         assert sig.expire_timestamp is None
         assert sig.primary_key_fingerprint == KEY_FINGERPRINT
     else:
+        assert len(sig) == 1
         assert sig.fingerprint == KEY_FINGERPRINT
         assert sig.timestamp == SIG_TIMESTAMP
         assert sig.expire_timestamp is None
@@ -457,6 +492,20 @@ def test_verify_manifest(openpgp_env, manifest_var, key_var, expected):
         pytest.skip(str(e))
 
 
+def test_verify_one_out_of_two():
+    try:
+        with MockedSystemGPGEnvironment() as openpgp_env:
+            with io.BytesIO(VALID_PUBLIC_KEY) as f:
+                openpgp_env.import_key(f)
+
+            with io.StringIO(TWO_SIGNATURE_MANIFEST) as f:
+                sig = openpgp_env.verify_file(f, require_all_good=False)
+
+            assert_signature(sig, "TWO_SIGNATURE_MANIFEST", expect_both=False)
+    except OpenPGPNoImplementation as e:
+        pytest.skip(str(e))
+
+
 def test_verify_untrusted_key():
     try:
         with MockedSystemGPGEnvironment() as openpgp_env:
@@ -547,30 +596,40 @@ def test_recursive_manifest_loader(tmp_path, openpgp_env, filename,
         pytest.skip(str(e))
 
 
-@pytest.mark.parametrize('manifest_var,key_var,expected',
-                         [(m, k, e) for m, k, e in MANIFEST_VARIANTS
-                          if k is not None])
-def test_cli(tmp_path, caplog, manifest_var, key_var, expected):
-    """Test Manifest verification via CLI"""
-    with open(tmp_path / '.key.bin', 'wb') as f:
-        f.write(globals()[key_var])
-    with open(tmp_path / 'Manifest', 'w') as f:
-        f.write(globals()[manifest_var])
+@pytest.fixture
+def base_tree(tmp_path):
     os.mkdir(tmp_path / 'eclass')
     with open(tmp_path / 'eclass' / 'Manifest', 'w'):
         pass
-    with open(tmp_path / 'myebuild-0.ebuild', 'wb') as f:
-        if manifest_var == 'MODIFIED_SIGNED_MANIFEST':
-            f.write(b'12345678901234567890123456789012')
+    with open(tmp_path / 'myebuild-0.ebuild', 'wb'):
+        pass
     with open(tmp_path / 'metadata.xml', 'wb'):
         pass
+    return tmp_path
+
+
+@pytest.mark.parametrize('manifest_var,key_var,expected',
+                         [(m, k, e) for m, k, e in MANIFEST_VARIANTS
+                          if k is not None])
+def test_cli(base_tree, caplog, manifest_var, key_var, expected):
+    """Test Manifest verification via CLI"""
+    with open(base_tree / '.key.bin', 'wb') as f:
+        f.write(globals()[key_var])
+    with open(base_tree / 'Manifest', 'w') as f:
+        f.write(globals()[manifest_var])
+    if manifest_var == 'MODIFIED_SIGNED_MANIFEST':
+        with open(base_tree / 'myebuild-0.ebuild', 'wb') as f:
+            f.write(b'12345678901234567890123456789012')
 
     retval = gemato.cli.main(['gemato', 'verify',
                               '--openpgp-key',
-                              str(tmp_path / '.key.bin'),
+                              str(base_tree / '.key.bin'),
                               '--no-refresh-keys',
                               '--require-signed-manifest',
-                              str(tmp_path)])
+                              # we verify this option separately
+                              # and our test data currently sucks
+                              '--no-require-secure-hashes',
+                              str(base_tree)])
     if str(OpenPGPNoImplementation('install gpg')) in caplog.text:
         pytest.skip('OpenPGP implementation missing')
 
@@ -633,7 +692,7 @@ def privkey_env(request):
     env.close()
 
 
-TEST_STRING = u'The quick brown fox jumps over the lazy dog'
+TEST_STRING = 'The quick brown fox jumps over the lazy dog'
 
 
 @pytest.mark.parametrize('keyid', [None, PRIVATE_KEY_ID])
@@ -711,7 +770,7 @@ def test_recursive_manifest_loader_save_submanifest(tmp_path, privkey_env):
     m.save_manifest('eclass/Manifest')
 
     m2 = ManifestFile()
-    with open(tmp_path / 'eclass' / 'Manifest', 'r') as f:
+    with open(tmp_path / 'eclass' / 'Manifest') as f:
         m2.load(f, openpgp_env=privkey_env)
     assert not m2.openpgp_signed
     assert m2.openpgp_signature is None
@@ -871,22 +930,26 @@ def test_refresh_wkd(openpgp_env_with_refresh,
             pytest.skip(str(e))
 
 
+@pytest.mark.parametrize('status', [401, 404, 500, ConnectionError])
 def test_refresh_wkd_fallback_to_hkp(openpgp_env_with_refresh,
-                                     hkp_server):
+                                     hkp_server, caplog, status):
     """Test whether WKD refresh failure falls back to HKP"""
     with pytest.importorskip('responses').RequestsMock() as responses:
         try:
             with io.BytesIO(VALID_PUBLIC_KEY) as f:
                 openpgp_env_with_refresh.import_key(f)
             hkp_server.keys[KEY_FINGERPRINT] = REVOKED_PUBLIC_KEY
-            responses.add(
-                responses.GET,
-                'https://example.com/.well-known/openpgpkey/hu/'
-                '5x66h616iaskmnadrm86ndo6xnxbxjxb?l=gemato',
-                status=404)
+            if status is not ConnectionError:
+                responses.add(
+                    responses.GET,
+                    'https://example.com/.well-known/openpgpkey/hu/'
+                    '5x66h616iaskmnadrm86ndo6xnxbxjxb?l=gemato',
+                    status=status)
 
+            caplog.set_level(logging.DEBUG)
             openpgp_env_with_refresh.refresh_keys(
                 allow_wkd=True, keyserver=hkp_server.addr)
+            assert 'failing due to failed request' in caplog.text
 
             with pytest.raises(OpenPGPRevokedKeyFailure):
                 with io.StringIO(SIGNED_MANIFEST) as f:
@@ -949,3 +1012,116 @@ def test_cli_gpg_wrap(tmp_path, caplog, command, expected, match):
     assert retval == expected
     if match is not None:
         assert match in caplog.text
+
+
+@pytest.mark.parametrize("hashes_arg,insecure", INSECURE_HASH_TESTS)
+@pytest.mark.parametrize(
+    "sign,require_secure",
+    [(None, None),
+     (False, None),
+     (True, None),
+     (None, False),
+     (True, False),
+     ])
+def test_recursive_manifest_loader_require_secure(tmp_path, privkey_env,
+                                                  hashes_arg, insecure,
+                                                  sign, require_secure):
+    with open(tmp_path / "Manifest", "w") as f:
+        f.write(SIGNED_MANIFEST)
+
+    ctx = (pytest.raises(ManifestInsecureHashes)
+           if insecure is not None and sign is not False
+           and require_secure is not False
+           else contextlib.nullcontext())
+    with ctx:
+        m = ManifestRecursiveLoader(tmp_path / "Manifest",
+                                    hashes=hashes_arg.split(),
+                                    require_secure_hashes=require_secure,
+                                    verify_openpgp=not sign,
+                                    sign_openpgp=sign,
+                                    openpgp_env=privkey_env)
+        if not sign:
+            assert m.openpgp_signed
+
+
+@pytest.mark.parametrize("hashes_arg,insecure", INSECURE_HASH_TESTS)
+@pytest.mark.parametrize(
+    "sign,require_secure",
+    [("", ""),
+     ("--no-sign", ""),
+     ("--sign", ""),
+     ("", "--no-require-secure-hashes"),
+     ("--sign", "--no-require-secure-hashes"),
+     ])
+def test_update_require_secure_cli(base_tree, caplog, hashes_arg,
+                                   insecure, sign, require_secure):
+    with open(base_tree / ".key.bin", "wb") as keyf:
+        keyf.write(PRIVATE_KEY)
+    with open(base_tree / "Manifest", "w") as f:
+        f.write(SIGNED_MANIFEST)
+
+    retval = gemato.cli.main(["gemato", "update",
+                              "-K", str(base_tree / ".key.bin"),
+                              "--hashes", hashes_arg,
+                              str(base_tree)]
+                             + f"{sign} {require_secure}".split())
+    if str(OpenPGPNoImplementation('install gpg')) in caplog.text:
+        pytest.skip('OpenPGP implementation missing')
+
+    expected = (1 if insecure is not None and sign != "--no-sign"
+                and require_secure != "--no-require-secure-hashes"
+                else 0)
+    assert retval == expected
+    if expected == 1:
+        assert str(ManifestInsecureHashes(insecure)) in caplog.text
+
+
+@pytest.mark.parametrize(
+    "require_secure", ["", "--no-require-secure-hashes"])
+def test_verify_require_secure_cli(base_tree, caplog, require_secure):
+    with open(base_tree / ".key.bin", "wb") as keyf:
+        keyf.write(VALID_PUBLIC_KEY)
+    with open(base_tree / "Manifest", "w") as f:
+        f.write(SIGNED_MANIFEST)
+
+    retval = gemato.cli.main(["gemato", "verify",
+                              "--no-refresh-keys",
+                              "--require-signed-manifest",
+                              "-K", str(base_tree / ".key.bin"),
+                              str(base_tree)]
+                             + require_secure.split())
+    if str(OpenPGPNoImplementation('install gpg')) in caplog.text:
+        pytest.skip('OpenPGP implementation missing')
+
+    expected = (1 if require_secure != "--no-require-secure-hashes"
+                else 0)
+    assert retval == expected
+    if expected == 1:
+        assert str(ManifestInsecureHashes(["MD5"])) in caplog.text
+
+
+@pytest.mark.parametrize(
+    "key_var,two_sigs",
+    [("TWO_SIGNATURE_PUBLIC_KEYS", True),
+     ("VALID_PUBLIC_KEY", False),
+     ])
+def test_verify_detached(tmp_path, key_var, two_sigs):
+    try:
+        with MockedSystemGPGEnvironment() as openpgp_env:
+            with io.BytesIO(globals()[key_var]) as f:
+                openpgp_env.import_key(f)
+
+            with open(tmp_path / "data.bin", "wb") as f:
+                f.write(b"\r\n".join(COMMON_MANIFEST_TEXT.encode("utf8")
+                                     .splitlines()))
+            with open(tmp_path / "sig.bin", "wb") as f:
+                f.write(base64.b64decode(TWO_SIGNATURES))
+
+            sig = openpgp_env.verify_detached(
+                tmp_path / "sig.bin", tmp_path / "data.bin",
+                require_all_good=two_sigs)
+
+            assert_signature(sig, "TWO_SIGNATURE_MANIFEST",
+                             expect_both=two_sigs)
+    except OpenPGPNoImplementation as e:
+        pytest.skip(str(e))
