@@ -1,9 +1,9 @@
 # gemato: Recursive loader tests
-# vim:fileencoding=utf-8
-# (c) 2017-2020 Michał Górny
+# (c) 2017-2022 Michał Górny
 # Licensed under the terms of 2-clause BSD license
 
 import base64
+import contextlib
 import datetime
 import gzip
 import itertools
@@ -20,8 +20,10 @@ from gemato.exceptions import (
     ManifestIncompatibleEntry,
     ManifestCrossDevice,
     ManifestSymlinkLoop,
+    ManifestNoSupportedHashes,
+    ManifestInsecureHashes,
     )
-from gemato.manifest import ManifestPathEntry
+from gemato.manifest import ManifestPathEntry, ManifestFileEntry
 from gemato.recursiveloader import ManifestRecursiveLoader
 
 from tests.test_compression import COMPRESSION_ALGOS
@@ -630,7 +632,7 @@ MD5 d41d8cd98f00b204e9800998ecf8427e
         FILENAME: '',
     }
 
-    MANIFESTS_REWRITTEN = dict((k, v.lstrip()) for k, v in MANIFESTS.items())
+    MANIFESTS_REWRITTEN = {k: v.lstrip() for k, v in MANIFESTS.items()}
 
 
 class FilenameBackslashLayout(BaseLayout):
@@ -644,7 +646,7 @@ DATA foo\\x5Cbar 0 MD5 d41d8cd98f00b204e9800998ecf8427e
         FILENAME: '',
     }
 
-    MANIFESTS_REWRITTEN = dict((k, v.lstrip()) for k, v in MANIFESTS.items())
+    MANIFESTS_REWRITTEN = {k: v.lstrip() for k, v in MANIFESTS.items()}
 
 
 class NewManifestLayout(BaseLayout):
@@ -788,6 +790,46 @@ DATA test 11 SHA1 561295c9cbf9d6b2f6428414504a8deed3020641
                 f'TIMESTAMP {future_dt.strftime("%Y-%m-%dT%H:%M:%SZ")}'))
 
 
+class UnknownHashLayout(BaseLayout):
+    """Layout with a supported and unknown hash"""
+
+    MANIFESTS = {
+        'Manifest': '''
+DATA test 0 MD5 d41d8cd98f00b204e9800998ecf8427e X-UNKNOWN 0123456789abcdef
+''',
+    }
+    FILES = {
+        'test': '',
+    }
+
+
+class UnknownHashOnlyLayout(BaseLayout):
+    """Layout with unknown hash only"""
+
+    MANIFESTS = {
+        'Manifest': '''
+DATA test 0 X-UNKNOWN 0123456789abcdef
+''',
+    }
+    FILES = {
+        'test': '',
+    }
+
+
+class SecureHashLayout(BaseLayout):
+    """Layout using at least one cryptographically secure hash"""
+
+    MANIFESTS = {
+        "Manifest": """
+DATA test 0 MD5 d41d8cd98f00b204e9800998ecf8427e\
+ SHA256 e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855
+""",
+    }
+    FILES = {
+        "test": "",
+    }
+
+
 FLAT_LAYOUTS = [
     DuplicateEntryLayout,
     DuplicateEbuildEntryLayout,
@@ -818,6 +860,8 @@ FLAT_LAYOUTS = [
     SymlinkLoopLayout,
     SymlinkLoopIgnoreLayout,
     MismatchedFileLayout,
+    UnknownHashLayout,
+    UnknownHashOnlyLayout,
 ]
 SUB_LAYOUTS = [
     SubTimestampLayout,
@@ -1069,6 +1113,7 @@ COMMON_VERIFY_PATH_VARIANTS = [
      [('MD5',
        '5f8db599de986fab7a21625b7916589c',
        '6f8db599de986fab7a21625b7916589c')]),
+    (UnknownHashLayout, 'test', True, []),
 ]
 
 
@@ -1151,6 +1196,8 @@ def test_assert_path_verifies(layout_factory, layout, path, expected, diff):
       {'': {'metadata.xml': ('DATA', 'metadata.xml', ['MD5'])}}),
      (IncompatibleTypeLayout, '', 'MISC',
       {'': {'metadata.xml': ('MISC', 'metadata.xml', ['MD5'])}}),
+     (UnknownHashLayout, '', None,
+      {'': {'test': ('DATA', 'test', ['MD5', 'X-UNKNOWN'])}}),
      ])
 def test_get_file_entry_dict(layout_factory, layout, path, only_types,
                              expected):
@@ -1160,10 +1207,9 @@ def test_get_file_entry_dict(layout_factory, layout, path, only_types,
     if only_types is not None:
         only_types = [only_types]
     entries = m.get_file_entry_dict(path, only_types=only_types)
-    assert (dict((subdir, dict((k, get_entry(v))
-                               for k, v in files.items()))
-                 for subdir, files in entries.items()) ==
-            expected)
+    assert ({subdir: {k: get_entry(v) for k, v in files.items()}
+             for subdir, files in entries.items()
+             } == expected)
 
 
 @pytest.mark.parametrize(
@@ -1227,8 +1273,8 @@ def test_get_deduplicated_file_entry_dict_for_update(layout_factory,
     m = ManifestRecursiveLoader(tmp_path / layout.TOP_MANIFEST,
                                 allow_xdev=False)
     entries = m.get_deduplicated_file_entry_dict_for_update(path)
-    assert dict((k, (v[0],) + get_entry(v[1]))
-                for k, v in entries.items()) == expected
+    assert {k: (v[0],) + get_entry(v[1])
+            for k, v in entries.items()} == expected
 
 
 @pytest.mark.parametrize(
@@ -1281,6 +1327,7 @@ COMMON_DIRECTORY_VERIFICATION_VARIANTS = [
      [('MD5',
        '5f8db599de986fab7a21625b7916589c',
        '6f8db599de986fab7a21625b7916589c')]),
+    (UnknownHashLayout, '', None, []),
 ]
 
 
@@ -1323,6 +1370,7 @@ COMMON_DIRECTORY_VERIFICATION_VARIANTS = [
      (SymlinkLoopLayout, '', lambda e: False, ManifestSymlinkLoop, None,
       []),
      (SymlinkLoopIgnoreLayout, '', None, True, None, []),
+     (UnknownHashOnlyLayout, '', None, ManifestNoSupportedHashes, None, []),
      ])
 def test_assert_directory_verifies(layout_factory, layout, path, fail_handler,
                                    expected, fail_path, diff):
@@ -1380,6 +1428,9 @@ def test_assert_directory_verifies(layout_factory, layout, path, fail_handler,
      (SymlinkLoopLayout, '', '',
       str(ManifestSymlinkLoop('<path>')).split('<path>', 1)[1]),
      (SymlinkLoopIgnoreLayout, '', '', None),
+     (UnknownHashOnlyLayout, '', '',
+      str(ManifestNoSupportedHashes("<path>", ManifestFileEntry(
+            "", 0, {"X-UNKNOWN": ""}))).split("<path>", 1)[1]),
      ])
 def test_cli_verify(layout_factory, caplog, layout, path, args, expected):
     tmp_path = layout_factory.create(layout, readonly=True)
@@ -1795,6 +1846,26 @@ def test_update_entry_for_path_no_hash_specified(layout_factory):
       {'Manifest':
        'MANIFEST a/Manifest 50 MD5 0f7cd9ed779a4844f98d28315dd9176a\n',
        }),
+     (SubManifestMismatchLayout,
+      ["MD5"],
+      ManifestRecursiveLoader.update_entry_for_path,
+      "a/Manifest",
+      None,
+      None,
+      {"Manifest":
+       "MANIFEST a/Manifest 50 MD5 0f7cd9ed779a4844f98d28315dd9176a\n",
+       }),
+     (SubManifestMismatchLayout,
+      ["MD5"],
+      ManifestRecursiveLoader.update_entry_for_path,
+      "a/test",
+      None,
+      None,
+      {"Manifest":
+       "MANIFEST a/Manifest 49 MD5 b86a7748346d54c6455886306f017e6c\n",
+       "a/Manifest":
+       "DATA test 0 MD5 d41d8cd98f00b204e9800998ecf8427e\n",
+       }),
      (NonexistingDirectoryLayout,
       ['MD5'],
       ManifestRecursiveLoader.update_entry_for_path,
@@ -1941,6 +2012,19 @@ def test_update_entry_raise(layout_factory, layout, path, expected, reason):
         assert exc.value.detail == reason
 
 
+@pytest.mark.parametrize(
+    'layout,path',
+    [(UnknownHashLayout, 'test'),
+     (UnknownHashOnlyLayout, 'test'),
+     ])
+def test_update_entry_unknown_hash(layout_factory, layout, path):
+    tmp_path = layout_factory.create(layout)
+    m = ManifestRecursiveLoader(tmp_path / layout.TOP_MANIFEST,
+                                allow_xdev=False)
+    with pytest.raises(KeyError):
+        m.update_entry_for_path(path)
+
+
 @pytest.mark.parametrize(
     'layout,path,expected',
     [(BasicTestLayout, 'nonexist', FileNotFoundError),
@@ -1970,7 +2054,7 @@ def test_update_entry_new_aux(layout_factory):
     assert (get_entry(m.find_path_entry('files/test.patch')) ==
             ('AUX', 'files/test.patch', ['MD5']))
     m.save_manifests()
-    with open(tmp_path / layout.TOP_MANIFEST, 'r') as f:
+    with open(tmp_path / layout.TOP_MANIFEST) as f:
         contents = f.read()
     assert (contents ==
             'AUX test.patch 0 MD5 d41d8cd98f00b204e9800998ecf8427e\n')
@@ -2069,7 +2153,7 @@ def test_cli_update(layout_factory, layout, args, update,
 
     if replace_timestamp is not None:
         m = gemato.manifest.ManifestFile()
-        with open(tmp_path / layout.TOP_MANIFEST, 'r') as f:
+        with open(tmp_path / layout.TOP_MANIFEST) as f:
             m.load(f)
         ts = m.find_timestamp()
         assert ts is not None
@@ -2276,7 +2360,7 @@ def test_new_manifest_cli(layout_factory, args):
     assert gemato.cli.main(['gemato', 'create', '--hashes=MD5'] +
                            args.split() + [str(tmp_path)]) == 0
 
-    with open(tmp_path / 'Manifest', 'r') as f:
+    with open(tmp_path / 'Manifest') as f:
         contents = f.read()
     expected = '''
 DATA test 0 MD5 d41d8cd98f00b204e9800998ecf8427e
@@ -2407,3 +2491,117 @@ def test_update_mtime(layout_factory, last_mtime, manifest_update):
     expected = dict(layout.MANIFESTS)
     expected.update(manifest_update)
     assert output == expected
+
+
+INSECURE_HASH_TESTS = [
+    # hashes, insecure
+    ("MD5", ["MD5"]),
+    ("SHA1", ["SHA1"]),
+    ("SHA512", None),
+    ("SHA1 SHA512", ["SHA1"]),
+    ("", []),
+]
+
+INSECURE_HASH_VERIFY_TESTS = [
+    # layout, insecure
+    (UnknownHashLayout, ["MD5"]),
+    (SecureHashLayout, None),
+]
+
+
+@pytest.mark.parametrize("hashes_arg,insecure", INSECURE_HASH_TESTS)
+def test_insecure_hashes(layout_factory, hashes_arg, insecure):
+    layout = BasicTestLayout
+    tmp_path = layout_factory.create(layout)
+    ctx = (pytest.raises(ManifestInsecureHashes) if insecure is not None
+           else contextlib.nullcontext())
+    with ctx:
+        ManifestRecursiveLoader(tmp_path / layout.TOP_MANIFEST,
+                                hashes=hashes_arg.split(),
+                                allow_xdev=False,
+                                require_secure_hashes=True)
+
+
+@pytest.mark.parametrize("layout,insecure", INSECURE_HASH_VERIFY_TESTS)
+@pytest.mark.parametrize(
+    "func,path",
+    [(ManifestRecursiveLoader.verify_path, "test"),
+     (ManifestRecursiveLoader.assert_path_verifies, "test"),
+     (ManifestRecursiveLoader.assert_directory_verifies, ""),
+     ])
+def test_insecure_hashes_verify(layout_factory, layout, insecure, func, path):
+    tmp_path = layout_factory.create(layout)
+    m = ManifestRecursiveLoader(tmp_path / layout.TOP_MANIFEST,
+                                allow_xdev=False,
+                                require_secure_hashes=True)
+
+    ctx = (pytest.raises(ManifestInsecureHashes) if insecure is not None
+           else contextlib.nullcontext())
+    with ctx:
+        func(m, path)
+
+
+def test_insecure_hashes_load(layout_factory):
+    layout = BasicTestLayout
+    tmp_path = layout_factory.create(layout)
+    m = ManifestRecursiveLoader(tmp_path / layout.TOP_MANIFEST,
+                                allow_xdev=False,
+                                require_secure_hashes=True)
+
+    with pytest.raises(ManifestInsecureHashes):
+        m.load_manifests_for_path("sub")
+
+
+@pytest.mark.parametrize("hashes_arg,insecure", INSECURE_HASH_TESTS)
+@pytest.mark.parametrize(
+    "func,arg",
+    [(ManifestRecursiveLoader.update_entry_for_path, "sub/deeper/test"),
+     (ManifestRecursiveLoader.update_entries_for_directory, "sub/deeper"),
+     ])
+def test_insecure_hashes_update(layout_factory, hashes_arg, insecure, func,
+                                arg):
+    layout = BasicTestLayout
+    tmp_path = layout_factory.create(layout)
+    m = ManifestRecursiveLoader(tmp_path / layout.TOP_MANIFEST,
+                                hashes=["SHA512"],
+                                allow_xdev=False,
+                                require_secure_hashes=True)
+    ctx = (pytest.raises(ManifestInsecureHashes) if insecure is not None
+           else contextlib.nullcontext())
+    with ctx:
+        func(m, arg, hashes=hashes_arg.split())
+
+
+def test_insecure_hashes_update_no_arg(layout_factory):
+    layout = BasicTestLayout
+    tmp_path = layout_factory.create(layout)
+    m = ManifestRecursiveLoader(tmp_path / layout.TOP_MANIFEST,
+                                allow_xdev=False,
+                                require_secure_hashes=True)
+    with pytest.raises(ManifestInsecureHashes):
+        m.update_entry_for_path("sub/deeper/test")
+
+
+@pytest.mark.parametrize("layout,insecure", INSECURE_HASH_VERIFY_TESTS)
+def test_insecure_hashes_verify_cli(layout_factory, caplog, layout,
+                                    insecure):
+    tmp_path = layout_factory.create(layout)
+    expected = 1 if insecure is not None else 0
+    assert gemato.cli.main(["gemato", "verify", "--require-secure-hashes",
+                            str(tmp_path)]) == expected
+    if insecure is not None:
+        assert str(ManifestInsecureHashes(insecure)) in caplog.text
+
+
+@pytest.mark.parametrize("hashes_arg,insecure", INSECURE_HASH_TESTS)
+@pytest.mark.parametrize("command", ["create", "update"])
+def test_insecure_hashes_update_cli(layout_factory, caplog,
+                                    hashes_arg, insecure, command):
+    layout = BasicTestLayout
+    tmp_path = layout_factory.create(layout)
+    expected = 1 if insecure is not None else 0
+    assert gemato.cli.main(["gemato", command, "--hashes", hashes_arg,
+                            "--require-secure-hashes", "--force-rewrite",
+                            str(tmp_path)]) == expected
+    if insecure:
+        assert str(ManifestInsecureHashes(insecure)) in caplog.text
