@@ -1,19 +1,22 @@
 # gemato: Verification tests
-# vim:fileencoding=utf-8
-# (c) 2017-2020 Michał Górny
+# (c) 2017-2022 Michał Górny
 # Licensed under the terms of 2-clause BSD license
 
+import contextlib
+import errno
 import itertools
 import os
 import os.path
 import socket
 import stat
+import unittest.mock
 
 import pytest
 
 from gemato.exceptions import (
     ManifestInvalidPath,
     ManifestCrossDevice,
+    ManifestInsecureHashes,
     )
 from gemato.hash import hash_path
 from gemato.manifest import new_manifest_entry
@@ -103,6 +106,12 @@ def get_checksums(path):
     }
 
 
+EMPTY_CHECKSUMS = {
+    "MD5": "d41d8cd98f00b204e9800998ecf8427e",
+    "SHA1": "da39a3ee5e6b4b0d3255bfef95601890afd80709",
+    "SHA256": "e3b0c44298fc1c149afbf4c8996fb924"
+              "27ae41e4649b934ca495991b7852b855",
+}
 TEST_PATH_SIZES = {
     'empty-file': 0,
     'regular-file': 43,
@@ -110,8 +119,8 @@ TEST_PATH_SIZES = {
     '/proc/version': 0,
 }
 TEST_PATH_CHECKSUMS = {
-    'empty-file': {'MD5': 'd41d8cd98f00b204e9800998ecf8427e',
-                   'SHA1': 'da39a3ee5e6b4b0d3255bfef95601890afd80709',
+    'empty-file': {'MD5': EMPTY_CHECKSUMS["MD5"],
+                   'SHA1': EMPTY_CHECKSUMS["SHA1"],
                    '__size__': TEST_PATH_SIZES['empty-file'],
                    },
     'regular-file': {'MD5': '9e107d9d372bb6826bd81d3542a419d6',
@@ -443,6 +452,53 @@ def test_unreadable_file(test_tree, function, args):
             pass
 
 
+@pytest.mark.parametrize(
+    "entry_hash,hashes_arg,insecure",
+    [("MD5", None, True),
+     ("SHA1", None, True),
+     ("SHA512", None, False),
+     ("MD5", "SHA1 SHA512", True),
+     ("MD5", "SHA512", False),
+     ("SHA512", "MD5 SHA512", True),
+     ("", None, True),
+     ("", "SHA512", False),
+     ("SHA512", "", True),
+     ])
+def test_insecure_hashes(test_tree, entry_hash, hashes_arg, insecure):
+    ctx = (pytest.raises(ManifestInsecureHashes) if insecure
+           else contextlib.nullcontext())
+    with ctx:
+        update_entry_for_path(
+            test_tree / "empty-file",
+            new_manifest_entry("DATA", "empty-file", 0,
+                               {entry_hash: ""} if entry_hash else {}),
+            hashes=hashes_arg.split() if hashes_arg is not None else None,
+            require_secure_hashes=True)
+
+
+@pytest.mark.parametrize(
+    "entry_hash,insecure",
+    [("", True),
+     ("MD5", True),
+     ("SHA1", True),
+     ("SHA256", False),
+     ("MD5 SHA1", True),
+     ("SHA1 SHA256", False),
+     ])
+def test_verify_insecure_hashes(test_tree, entry_hash, insecure):
+    checksums = {}
+    for h in entry_hash.split():
+        checksums[h] = EMPTY_CHECKSUMS[h]
+
+    ctx = (pytest.raises(ManifestInsecureHashes) if insecure
+           else contextlib.nullcontext())
+    with ctx:
+        verify_path(
+            test_tree / "empty-file",
+            new_manifest_entry("DATA", "empty-file", 0, checksums),
+            require_secure_hash=True)
+
+
 @pytest.mark.parametrize(
     'a_cls,a_name,a_args,b_cls,b_name,b_args,expected,diff',
     [('DATA', 'test', [0, {'MD5': 'd41d8cd98f00b204e9800998ecf8427e'}],
@@ -505,3 +561,11 @@ def test_entry_compatibility(a_cls, a_name, a_args, b_cls, b_name,
     e1 = new_manifest_entry(a_cls, a_name, *a_args)
     e2 = new_manifest_entry(b_cls, b_name, *b_args)
     assert verify_entry_compatibility(e1, e2) == (expected, diff)
+
+
+def test_get_file_metadata_fail_to_open_reg(test_tree):
+    """Regression test for when open() fails on a seemingly regular file"""
+    with unittest.mock.patch("os.open") as mock_open:
+        mock_open.side_effect = OSError(errno.ENXIO, "mocked error")
+        with pytest.raises(OSError):
+            list(get_file_metadata(test_tree / "regular-file", {}))
